
From: Dave Hansen <haveblue@us.ibm.com>

This splits up sparse_index_alloc() into two pieces.  This is needed
because we'll allocate the memory for the second level in a different place
from where we actually consume it to keep the allocation from happening
underneath a lock

Signed-off-by: Dave Hansen <haveblue@us.ibm.com>
Signed-off-by: Andrew Morton <akpm@osdl.org>
---

 include/linux/mmzone.h |    1 
 mm/sparse.c            |   56 +++++++++++++++++++++++++++++++++++--------------
 2 files changed, 42 insertions(+), 15 deletions(-)

diff -puN include/linux/mmzone.h~sparsemem-extreme-hotplug-preparation include/linux/mmzone.h
--- 25/include/linux/mmzone.h~sparsemem-extreme-hotplug-preparation	Wed Aug 17 14:20:29 2005
+++ 25-akpm/include/linux/mmzone.h	Wed Aug 17 14:20:29 2005
@@ -588,6 +588,7 @@ static inline int pfn_valid(unsigned lon
 void sparse_init(void);
 #else
 #define sparse_init()	do {} while (0)
+#define sparse_index_init(_sec, _nid)  do {} while (0)
 #endif /* CONFIG_SPARSEMEM */
 
 #ifdef CONFIG_NODES_SPAN_OTHER_NODES
diff -puN mm/sparse.c~sparsemem-extreme-hotplug-preparation mm/sparse.c
--- 25/mm/sparse.c~sparsemem-extreme-hotplug-preparation	Wed Aug 17 14:20:29 2005
+++ 25-akpm/mm/sparse.c	Wed Aug 17 14:20:29 2005
@@ -6,6 +6,7 @@
 #include <linux/mmzone.h>
 #include <linux/bootmem.h>
 #include <linux/module.h>
+#include <linux/spinlock.h>
 #include <asm/dma.h>
 
 /*
@@ -22,27 +23,52 @@ struct mem_section mem_section[NR_SECTIO
 #endif
 EXPORT_SYMBOL(mem_section);
 
-static void sparse_alloc_root(unsigned long root, int nid)
-{
 #ifdef CONFIG_SPARSEMEM_EXTREME
-	mem_section[root] = alloc_bootmem_node(NODE_DATA(nid), PAGE_SIZE);
-#endif
-}
-
-static void sparse_index_init(unsigned long section, int nid)
+static struct mem_section *sparse_index_alloc(int nid)
 {
-	unsigned long root = SECTION_NR_TO_ROOT(section);
+	struct mem_section *section = NULL;
+	unsigned long array_size = SECTIONS_PER_ROOT *
+				   sizeof(struct mem_section);
 
-	if (mem_section[root])
-		return;
+	section = alloc_bootmem_node(NODE_DATA(nid), array_size);
 
-	sparse_alloc_root(root, nid);
+	if (section)
+		memset(section, 0, array_size);
 
-	if (mem_section[root])
-		memset(mem_section[root], 0, PAGE_SIZE);
-	else
-		panic("memory_present: NO MEMORY\n");
+	return section;
 }
+
+static int sparse_index_init(unsigned long section_nr, int nid)
+{
+	static spinlock_t index_init_lock = SPIN_LOCK_UNLOCKED;
+	unsigned long root = SECTION_NR_TO_ROOT(section_nr);
+	struct mem_section *section;
+	int ret = 0;
+
+	section = sparse_index_alloc(nid);
+	/*
+	 * This lock keeps two different sections from
+	 * reallocating for the same index
+	 */
+	spin_lock(&index_init_lock);
+
+	if (mem_section[root]) {
+		ret = -EEXIST;
+		goto out;
+	}
+
+	mem_section[root] = section;
+out:
+	spin_unlock(&index_init_lock);
+	return ret;
+}
+#else /* !SPARSEMEM_EXTREME */
+static inline int sparse_index_init(unsigned long section_nr, int nid)
+{
+	return 0;
+}
+#endif
+
 /* Record a memory area against a node. */
 void memory_present(int nid, unsigned long start, unsigned long end)
 {
_
